@@ -95,7 +95,6 @@ typedef struct _State {
 } State;
 
 static State *all_instances = NULL;
-static State global = { { { { -1, -1 }, NULL, }, }, NULL, NULL, NULL, FIRST_HANDLE, NULL };
 
 #define PROXY_VALID(px) ((px) && (px)->forkid == p11_forkid)
 #define PROXY_FORKED(px) ((px) && (px)->forkid != p11_forkid)
@@ -250,8 +249,7 @@ modules_dup (CK_FUNCTION_LIST **modules)
 }
 
 static CK_RV
-proxy_create (Proxy **res, CK_FUNCTION_LIST **loaded,
-	      Mapping *mappings, unsigned int n_mappings)
+proxy_list_slots (Proxy *py, Mapping *mappings, unsigned int n_mappings)
 {
 	CK_FUNCTION_LIST_PTR *f;
 	CK_FUNCTION_LIST_PTR funcs;
@@ -259,6 +257,58 @@ proxy_create (Proxy **res, CK_FUNCTION_LIST **loaded,
 	CK_ULONG i, count;
 	unsigned int j;
 	CK_RV rv = CKR_OK;
+
+	for (f = py->inited; *f; ++f) {
+		funcs = *f;
+		assert (funcs != NULL);
+		slots = NULL;
+
+		/* Ask module for its slots */
+		rv = (funcs->C_GetSlotList) (FALSE, NULL, &count);
+		if (rv == CKR_OK && count) {
+			slots = calloc (sizeof (CK_SLOT_ID), count);
+			rv = (funcs->C_GetSlotList) (FALSE, slots, &count);
+		}
+
+		if (rv != CKR_OK) {
+			free (slots);
+			break;
+		}
+
+		return_val_if_fail (count == 0 || slots != NULL, CKR_GENERAL_ERROR);
+
+		if (count > 0) {
+			py->mappings = realloc (py->mappings, sizeof (Mapping) * (py->n_mappings + count));
+			return_val_if_fail (py->mappings != NULL, CKR_HOST_MEMORY);
+
+			/* And now add a mapping for each of those slots */
+			for (i = 0; i < count; ++i) {
+				/* Reuse the existing mapping if any */
+				for (j = 0; j < n_mappings; ++j) {
+					if (mappings[j].funcs == funcs &&
+					    mappings[j].real_slot == slots[i])
+						break;
+				}
+				py->mappings[py->n_mappings].funcs = funcs;
+				py->mappings[py->n_mappings].wrap_slot =
+					(n_mappings == 0 || j == n_mappings) ?
+					py->n_mappings + MAPPING_OFFSET :
+					mappings[j].wrap_slot;
+				py->mappings[py->n_mappings].real_slot = slots[i];
+				++py->n_mappings;
+			}
+		}
+
+		free (slots);
+	}
+	return rv;
+}
+
+static CK_RV
+proxy_create (Proxy **res, CK_FUNCTION_LIST **loaded,
+	      Mapping *mappings, unsigned int n_mappings)
+{
+	CK_RV rv = CKR_OK;
 	Proxy *py;
 
 	py = calloc (1, sizeof (Proxy));
@@ -267,51 +317,15 @@ proxy_create (Proxy **res, CK_FUNCTION_LIST **loaded,
 	py->forkid = p11_forkid;
 
 	py->inited = modules_dup (loaded);
-	return_val_if_fail (py->inited != NULL, CKR_HOST_MEMORY);
+	if (py->inited == NULL) {
+		proxy_free (py, 0);
+		return_val_if_reached (CKR_HOST_MEMORY);
+	}
 
 	rv = p11_kit_modules_initialize (py->inited, NULL);
 
 	if (rv == CKR_OK) {
-		for (f = py->inited; *f; ++f) {
-			funcs = *f;
-			assert (funcs != NULL);
-			slots = NULL;
-
-			/* Ask module for its slots */
-			rv = (funcs->C_GetSlotList) (FALSE, NULL, &count);
-			if (rv == CKR_OK && count) {
-				slots = calloc (sizeof (CK_SLOT_ID), count);
-				rv = (funcs->C_GetSlotList) (FALSE, slots, &count);
-			}
-
-			if (rv != CKR_OK) {
-				free (slots);
-				break;
-			}
-
-			return_val_if_fail (count == 0 || slots != NULL, CKR_GENERAL_ERROR);
-
-			if (count > 0) {
-				py->mappings = realloc (py->mappings, sizeof (Mapping) * (py->n_mappings + count));
-				return_val_if_fail (py->mappings != NULL, CKR_HOST_MEMORY);
-
-				/* And now add a mapping for each of those slots */
-				for (i = 0; i < count; ++i) {
-					/* Reuse the existing mapping if any */
-					for (j = 0; j < n_mappings; ++j) {
-						if (mappings[j].funcs == funcs &&
-						    mappings[j].real_slot == slots[i])
-							break;
-					}
-					py->mappings[py->n_mappings].funcs = funcs;
-					py->mappings[py->n_mappings].wrap_slot = j == n_mappings ? py->n_mappings + MAPPING_OFFSET : mappings[j].wrap_slot;
-					py->mappings[py->n_mappings].real_slot = slots[i];
-					++py->n_mappings;
-				}
-			}
-
-			free (slots);
-		}
+		rv = proxy_list_slots (py, mappings, n_mappings);
 	}
 
 	if (rv != CKR_OK) {
@@ -320,7 +334,10 @@ proxy_create (Proxy **res, CK_FUNCTION_LIST **loaded,
 	}
 
 	py->sessions = p11_dict_new (p11_dict_ulongptr_hash, p11_dict_ulongptr_equal, NULL, free);
-	return_val_if_fail (py->sessions != NULL, CKR_HOST_MEMORY);
+	if (py->sessions == NULL) {
+		proxy_free (py, 1);
+		return_val_if_reached (CKR_HOST_MEMORY);
+	}
 	py->refs = 1;
 
 	*res = py;
@@ -445,7 +462,29 @@ proxy_C_GetSlotList (CK_X_FUNCTION_LIST *self,
 
 		if (!PROXY_VALID (state->px)) {
 			rv = CKR_CRYPTOKI_NOT_INITIALIZED;
-		} else {
+		}
+
+		if (rv == CKR_OK) {
+			Mapping *mappings = NULL;
+			unsigned int n_mappings = 0;
+
+			if (state->px->mappings) {
+				mappings = state->px->mappings;
+				n_mappings = state->px->n_mappings;
+				state->px->mappings = NULL;
+				state->px->n_mappings = 0;
+			}
+			rv = proxy_list_slots (state->px, mappings, n_mappings);
+			if (rv == CKR_OK) {
+				free (mappings);
+			} else {
+				p11_debug ("failed to list slots: %lu", rv);
+				state->px->mappings = mappings;
+				state->px->n_mappings = n_mappings;
+			}
+		}
+
+		if (rv == CKR_OK) {
 			index = 0;
 
 			/* Go through and build up a map */
@@ -564,7 +603,40 @@ proxy_C_WaitForSlotEvent (CK_X_FUNCTION_LIST *self,
                           CK_SLOT_ID_PTR slot,
                           CK_VOID_PTR reserved)
 {
-	return CKR_FUNCTION_NOT_SUPPORTED;
+	State *state = (State *)self;
+	Proxy *py = state->px;
+	CK_FUNCTION_LIST_PTR *f;
+	CK_FUNCTION_LIST_PTR funcs;
+	CK_SLOT_ID real_slot;
+	unsigned int i;
+	CK_RV rv = CKR_NO_EVENT;
+
+	/* Only the non-blocking case is supported. */
+	if ((flags & CKF_DONT_BLOCK) == 0)
+		return CKR_FUNCTION_NOT_SUPPORTED;
+
+	p11_lock ();
+
+	for (f = py->inited; *f; ++f) {
+		funcs = *f;
+		assert (funcs != NULL);
+
+		rv = (funcs->C_WaitForSlotEvent) (flags, &real_slot, reserved);
+		if (rv == CKR_NO_EVENT)
+			continue;
+		if (rv != CKR_OK)
+			break;
+		for (i = 0; i < py->n_mappings; i++)
+			if (py->mappings[i].funcs == funcs &&
+			    py->mappings[i].real_slot == real_slot) {
+				*slot = py->mappings[i].wrap_slot;
+				break;
+			}
+	}
+
+	p11_unlock ();
+
+	return rv;
 }
 
 static CK_RV
@@ -606,7 +678,8 @@ proxy_C_OpenSession (CK_X_FUNCTION_LIST *self,
 				sess->wrap_slot = map.wrap_slot;
 				sess->real_session = *handle;
 				sess->wrap_session = ++state->last_handle; /* TODO: Handle wrapping, and then collisions */
-				p11_dict_set (state->px->sessions, &sess->wrap_session, sess);
+				if (!p11_dict_set (state->px->sessions, &sess->wrap_session, sess))
+					warn_if_reached ();
 				*handle = sess->wrap_session;
 			}
 
@@ -1576,14 +1649,6 @@ proxy_C_GenerateRandom (CK_X_FUNCTION_LIST *self,
 	return (map.funcs->C_GenerateRandom) (handle, random_data, random_len);
 }
 
-static CK_RV
-module_C_WaitForSlotEvent (CK_FLAGS flags,
-                           CK_SLOT_ID_PTR slot,
-                           CK_VOID_PTR reserved)
-{
-	return proxy_C_WaitForSlotEvent (&global.virt.funcs, flags, slot, reserved);
-}
-
 /* --------------------------------------------------------------------
  * MODULE ENTRY POINT
  */
@@ -1696,11 +1761,8 @@ C_GetFunctionList (CK_FUNCTION_LIST_PTR_PTR list)
 		}
 	}
 
-	if (rv == CKR_OK) {
-		/* We use this as a check below */
-		module->C_WaitForSlotEvent = module_C_WaitForSlotEvent;
+	if (rv == CKR_OK)
 		*list = module;
-	}
 
 	if (loaded)
 		p11_kit_modules_release (loaded);
@@ -1720,25 +1782,29 @@ p11_proxy_module_cleanup (void)
 
 	for (; state != NULL; state = next) {
 		next = state->next;
-		p11_virtual_unwrap (state->wrapped);
 		p11_kit_modules_release (state->loaded);
+		p11_virtual_unwrap (state->wrapped);
 	}
 }
 
 bool
 p11_proxy_module_check (CK_FUNCTION_LIST_PTR module)
 {
-	return (module->C_WaitForSlotEvent == module_C_WaitForSlotEvent);
-}
+	State *state;
+	bool ret = false;
 
-static void
-proxy_module_free (p11_virtual *virt)
-{
-	State *state = (State *)virt;
+	if (!p11_virtual_is_wrapper (module))
+		return false;
 
-	p11_virtual_unwrap (state->wrapped);
-	p11_kit_modules_release (state->loaded);
-	free (state);
+	p11_lock ();
+	for (state = all_instances; state != NULL; state = state->next)
+		if (state->wrapped == module) {
+			ret = true;
+			break;
+		}
+	p11_unlock ();
+
+	return ret;
 }
 
 CK_RV
@@ -1758,9 +1824,10 @@ p11_proxy_module_create (CK_FUNCTION_LIST_PTR *module,
 	p11_virtual_init (&state->virt, &proxy_functions, state, NULL);
 	state->last_handle = FIRST_HANDLE;
 	state->loaded = modules_dup (modules);
-	state->wrapped = p11_virtual_wrap (&state->virt, (p11_destroyer)proxy_module_free);
+	state->wrapped = p11_virtual_wrap (&state->virt, (p11_destroyer)p11_virtual_uninit);
 	if (state->wrapped == NULL) {
-		proxy_module_free (&state->virt);
+		p11_kit_modules_release (state->loaded);
+		free (state);
 		return CKR_GENERAL_ERROR;
 	}
 
